/*
   There are three kind of spaces in content:
   Generator (G), Mathematic (M), and Visible (V) spaces.
   1. G-space, a space in which a generator computes points.
      Curves in this space are called paths.
   2. M-space, a space to which each G-point is mapped.
      For each G-point there may be several M-points.
      The map G -> M x M ... x M is defined by generator-
      dependent decoder.
      Curves in this space are called traces.
   3. V-space, a space visible in windows and to which
      M-points are mapped by a set of user-defined functions.
      Curves in this space are called shifts.
*/

#include "visual0.h"

/* typedef enum {...} VisualSpecAction migrated to common.h */
Global(void) VisualizerSpecial(VisualSpecAction vsa);

/* Color and ColorPtr migrated to common.h */

extern Global(Uint4) defForeground;
extern Global(Uint4) defBackground;

#include "document.h"

extern Global(MenuDesc) WindowMenuDesc;

Global(void) BuildWindowsList(Int2 i);
Global(void) DestroyWindowsList(void);

Global(void) lInvertColors(void);
Global(void) llString(Int2 x, Int2 y, CharPtr str, float PNTR hardcopy);
Global(void) lPaintOval(RectPtr r);
Global(void) lFrameOval(RectPtr r);
Global(void) lPaintRect(RectPtr r);
Global(void) lFrameRect(RectPtr r);
Global(void) lPaintPoly(Int2 num, PointPtr poly);
Global(void) lFramePoly(Int2 num, PointPtr poly);

Global(void) lMoveTo(Int2 x, Int2 y);
Global(void) lPoint(Int2 x, Int2 y);
Global(void) lLine(Int2 x1, Int2 y1, Int2 x2, Int2 y2);
Global(void) lString(Int2 x, Int2 y, CharPtr str);

Global(void) SetCurrentPen(DrawAttrPtr attr);

Global(void) PalDrawProc(PaneL pnl);

Global(void) ClearAllWindows(Int2 index);
Global(Boolean) CreateWindow_Status(Int2 index);
Global(void) DeleteWindows(DataLibPtr arch);
Global(void) NotifyAboutNewInitPoint(void);

Global(void) SaveVisualAttr(DataLibPtr dl, ByteStorePtr from,  ByteStorePtr to, Boolean freemem);
Global(void) RestoreVisualAttr(DataLibPtr dl, ByteStorePtr from,  ByteStorePtr to);
Global(void) PushVisualAttr(Boolean freemem);
Global(void) PopVisualAttr(void);

Global(void) CreateButtons(WindoW w, GrouP gt);

Global(void) ReadPalette(void);
Global(void) FreePalette(void);
Global(void) ColorSet(Color index);
Global(Int2) ColorNumber(void);
Global(long) ColorOut(Color i);
Global(Color) ColorIn(long rgb);
Global(Color) ColorAsk(Color InitColor);

extern Global(FloatHiPtr PNTR PNTR) IndirectValues;

Global(void) CompileWindowsFunctions(Boolean UnlockOld);

Global(void) GetNameAndDim(Int2 class, Int2 var, CharPtr PNTR name, Int2Ptr dim);
Global(Int2Ptr) GetClassDim(CharPtr Class);
Global(Uint2) GetClassId(CharPtr Class);
Global(Int2) GetNthRelevantClass(Int2 Class);
Global(Boolean) GetClassNameAndId(Int2 Class, CharPtr PNTR name, Uint2Ptr id);
Global(void) ClassAddVar(CharPtr clas, CharPtr var);
Global(void) EnumClasses(LisT List, ProcClassPtr proc);
Global(void) EnumVariables(LisT List, Int2 Class, ProcVariablesPtr proc);
Global(Boolean) IsClassBasic(Int2 Class);
Global(Int2) IsClassRelevant(Uint2 realId);

typedef enum {EA_NO,EA_DO,EA_DONT} ExpandOp;
Global(CharPtr) ExpandClassDef(CharPtr p, ExpandOp ExpandArrays);

Global(void) ClassTableReset(void);
Global(Int2) ClassTableSetPtr(CharPtr clas, VoidPtr ptr, Uint2Ptr id);
Global(Int2) ClassTableMaxIrrelevantDim(void);
Global(void) ClassTableSetIrrelevant(VoidPtr ptr);
Global(void) ClassSaveIndirect(void);
Global(void) ClassRestoreIndirect(void);

extern Global(Int2) VCnum,VCsep;
extern Global(CharPtr PNTR) VCptr;

extern Global(Boolean) VisFull;
Global(void) ActivateVisualizer(DataLibPtr arch);
Global(void) DeactivateVisualizer(DataLibPtr arch);

Global(void) ReadVisibleNames(void);
Global(void) FreeVisibleNames(void);

#if DEB_VIS
  Global(void) SnapWindowDescriptors(Int2);
  Global(void) Debug_VisualList(Int2);
#endif

Global(void) VisualizerInitGCurve(void);
Global(void) VisualizerTermGCurve(void);

Global(void) VisualizerInitMCurve(void);
Global(void) VisualizerTermMCurve(void);
Global(void) VisualizerProcMPoint(void);

Global(void) VisualizerFreeCurve(FilePtr fp);

#define Par(string) GetParam(SFS_VISUAL,string)

extern Global(int) textwidth;		/* width of DialogText control to edit functions */
extern Global(int) numwidth;		/* width of DialogText control to edit numbers */
extern Global(int) lines,linelen;	/* dimensions of list box with visible names */

Global(void) ShowHiddenWindows(Int2 index);
Global(Boolean) ShowHiddenWindow_Status(Int2 index);

Global(void) PauseItemsAll(Boolean enable);


extern Global(Color) BlackIndex;
extern Global(Color) WhiteIndex;

Global(void) ReadPsFonts(Int2 num);
Global(void) FreePsFonts(void);

Global(void) WindowsExport(DataLibPtr arch, FILE PNTR out);
Global(void) WindowsImport(DataLibPtr arch, FILE PNTR in);
Global(void) ExportVisualAttr(DataLibPtr dgm, FILE PNTR out);
Global(void) ImportVisualAttr(DataLibPtr dgm, FILE PNTR in);

Global(void) VisualizerReevalGDim(void);

